UNIT Search;
{search and comparison routines}

INTERFACE

{check string for search data}
{return True if found, otherwise False}
function CompareBuf(SearchStr,CompBuf : String) : Boolean;

{process search strings within parenthesis}
procedure ProcessParens(var SearchStr : String; CompBuf : String);


IMPLEMENTATION

{convert strings to uppercase}
procedure StUpCase(var S : String);
var
   P : Integer;

begin
     for P := 1 to Length(S) do
         S[P] := UpCase(S[P])
end;

{remove spaces from strings}
function TrimSpaces(S : String) : String;
var
   Position     : Integer;   {current position in entered string}
   NewPosition  : Integer;   {current position in converted string}
   Len          : Integer;   {length of entered string}
   Temp         : String;    {converted string}
   Space        : String[1];

begin
     Space := ' ';
     Temp := '';
     NewPosition := 0;
     Len := Length(S);
     {remove spaces}
     for Position := 1 to Len do begin
         if (S[Position] <> Space) then begin
              Inc(NewPosition);
              Temp[NewPosition] := S[Position]
         end
     end;
     {set length of new string}
     Temp[0] := Chr(NewPosition);
     {function return}
     TrimSpaces := Temp
end;

procedure GetToken(var OpStr : Char;       {And or Or symbol}
                   var SearchStr : String; {data to search for}
                   var Token : String);    {data before, after, 
                                            or between symbols}
{by symbols is meant the | for Or and & for And between search 
 words or phrases}
var
  OpPos    : Integer; {symbol position}
  OpPos1   : Integer; {first symbol position if more than one 
                       is used}
  OpPos2   : Integer; {second symbol position if more than one 
                       is used}
  OpStr1   : Char;    {first symbol if more than one is used}
  OpStr2   : Char;    {second symbol if more than one is used}
  Found    : Boolean; {a symbol has been found}
  Position : Integer; {temporary symbol position value}
  Start    : Integer; {start position of current token in search 
                       string}

begin
  Found  := False; {no symbols found}
  OpStr1 := ' ';
  OpStr2 := ' ';
  OpPos1 := 255;   {highest possible value}
  OpPos2 := 255;
  Start  := 1;     {beginning of string}

  Position := Pos('|',SearchStr); {look for first occurance of Or
                                   in search string}
  if Position > 0 then begin      {found it}
    OpStr1 := '|';                {found what?}
    OpPos1 := Position;           {found it where?}
    Found := True
  end;

  Position := Pos('&',SearchStr); {look for first occurance of And
                                   in search string}

  if Position > 0 then begin      {found it}
    OpStr2 := '&';                {found what?}
    OpPos2 := Position;           {found it where?}
    Found := True
  end;

  if Found then begin              {found at least one symbol}
    if OpPos1 < OpPos2 then begin  {which symbol is first in search
                                    string}
      OpPos := OpPos1;             {symbol at position 1 was first}
      OpStr := OpStr1
    end
    else begin
      OpPos := OpPos2;             {symbol at position 2 was first}
      OpStr := OpStr2
    end;

    Token := Copy(SearchStr,Start,OpPos - Start); {retrieve the 
                                                   data before 
                                                   current symbol}

    {delete data just retrieved, including symbol, from search 
     string to make future searches simpler}
    Delete(SearchStr,Start,(OpPos - Start) + 1)
  end                          

  else begin                    {no symbol found}
    OpPos := Length(SearchStr); {position set to end of search 
                                 string}
    OpStr := ' ';               {no symbol found}
    Token := SearchStr;         {therefore use the entire search 
                                 string}
    SearchStr := ''             {delete search string to indicate 
                                 done}
  end;
  StUpCase(Token)               {convert token to upper-case. All 
                                 searches use upper case so they 
                                 won't be case-sensitive}
end;


function CompareBuf(SearchStr, CompBuf : String) : Boolean;
{symbols are And = '&', Or = '|'}
{search for token in record}
var
  NotFound   : Integer;   {string search for token unsuccessful}
  Token      : String;    {string to search for}
  Arg        : Integer;   {current token found (1 : True, 
                           0 : False)}
  LastArg    : Integer;   {previous token found (1 : True, 
                           0 : False)}
  OpStr      : Char;      {current symbol}
  OldOpStr   : Char;      {previous symbol}

begin
  StUpCase(CompBuf);       {make record text upper-case for
                            case-insensitive searching}

  CompBuf := TrimSpaces(CompBuf);  {delete spaces in text to search}
  NotFound := 0;
  GetToken(OpStr,SearchStr,Token); {get the current token and symbol 
                                    from current search string. As 
                                    tokens are extracted the search 
                                    string is shortened}
  Token := TrimSpaces(Token);      {get rid of any of token's 
                                    leading or trailing spaces. This 
                                    is very important as such spaces 
                                    are usually included in the 
                                    token, wfich could cause the 
                                    search to fail}
  if Token = '~1~' then
    LastArg := 1                      {set argument to 1 for True}
  else begin
    if Token = '~0~' then
      LastArg := 0                    {set argument to 0 for False}
    else begin
      if Pos(Token,CompBuf) = NotFound then {first token not found}
        LastArg := 0                  {set argument to 0 for False}
      else
        LastArg := 1                  {set argument to 1 for True}
    end
  end;

  if OpStr <> ' ' then begin          {no symbols found in search 
                                            string}
    repeat
      OldOpStr := OpStr;               {save current symbol}
      GetToken(OpStr,SearchStr,Token); {get new token and symbol}
      Token := TrimSpaces(Token);      {get rid of surrounding 
                                        spaces}
      case OldOpStr of
        '|' :                          {Or}
          begin
            if Token = '~1~' then      {compared result of search 
                                        data in parentheses}
              Arg := 1
            else begin
              if Token = '~0~' then
                Arg := 0
              else begin
                if Pos(Token,CompBuf) = NotFound then
                  Arg := 0             {set argument to 0 for False}
                else
                  Arg := 1             {set argument to 1 for True}
              end
            end;
            LastArg := LastArg or Arg  {OR current with previous 
                                        cumulative results}
          end;

        '&' :
          begin
            if Token = '~1~' then      {compared result of search
                                        data in parentheses}

              Arg := 1
            else begin
              if Token = '~0~' then
                Arg := 0
              else begin
                if Pos(Token,CompBuf) = NotFound then
                  Arg := 0             {set argument to 0 for False}
                else
                  Arg := 1             {set argument to 1 for True}
              end
            end;
            LastArg := LastArg and Arg {AND current with previous
                                        cumulative results}
          end
      end;
    until OpStr = ' '                  {search string completed}
  end;

  if LastArg = 1 then                  {search successful}
    CompareBuf := True                     {function return}
  else
    CompareBuf := False                    {function return}
end;

procedure ProcessParens(var SearchStr : String; CompBuf : String);
{process search strings within parenthesis}
const
  Left  : Char = '(';
  Right : Char = ')';
var
  Abort           : Boolean; {no parentheses or they are mismatched}
  Comparison      : String;  {result of comparison - ~0~ False, 
                              ~1~ True}
  Counter         : Integer; {for stepping through search string}
  FirstLeftParen  : Integer; {position of first left parentheses in
                              search string - used to be sure 
                              parentheses exist in string}
  FirstRightParen : Integer; {see above}
  LeftPos         : Integer; {position of rightmost left parenthese}
  RightPos        : Integer; {position of first right parentheses to
                              right of LeftPos}
  Size            : Integer; {size of search string}
  TempSearch      : String;  {contains search data from within 
                              parentheses}
                        

begin
  Abort := False;
  repeat
    Size := length(SearchStr);               {number of characters 
                                              in search string}
    FirstLeftParen := Pos(Left,SearchStr);   {is there at least one 
                                              left parentheses in 
                                              search string?}
    FirstRightParen := Pos(Right,SearchStr); {is there at least one 
                                              right parentheses in 
                                              search string?}

    if (FirstLeftParen = 0) or (FirstRightParen = 0) {one or no 
                                                      parentheses
                                                      found}
        or (FirstRightParen < FirstLeftParen) then   {mismatched 
                                                      parentheses}
      Abort := True                                  {can't process 
                                                      search string}
    else begin
      Counter := Size;
      {look for rightmost left parenthese}
      while (SearchStr[Counter] <> Left) and (Counter > 0) do
        Dec(Counter);
      LeftPos := Counter; {position of rightmost left parentheses}
      {now look for first right parentheses following it}
      while (SearchStr[Counter] <> Right) and (Counter <= Size) do
        Inc(Counter);
      RightPos := Counter; {position of first right parenthese
                            following rightmost left parentheses}
      if (RightPos > Size) or (LeftPos > RightPos) then
        Abort := True  {no right paren found after left}
      else begin
        {isolate contents of parentheses pair we found}
        TempSearch := copy(SearchStr,LeftPos + 1,RightPos
                                     - (LeftPos + 1));
        {found match in record?}
        if CompareBuf(TempSearch,CompBuf) then
          Comparison := ' ~1~ ' {yes}
        else
          Comparison := ' ~0~ '; {no}
        {tried to use something that user would be unlikely to
         include in the search string}
        {delete parentheses and contents from search string}
        Delete(SearchStr,LeftPos,(RightPos - LeftPos) + 1);
        {replace with ~0~ or ~1~}
        Insert(Comparison,SearchStr,LeftPos)
      end
    end;
  until Abort {repeat until we run out of parentheses pairs}
end; {end of procedure}

end. {end of unit}
